# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import codecs

__all__ = ['ensure_utf8', 'u', 'write_utf8_to']


# if python 3.x
if sys.version_info[0] == 2:
    def write_utf8_to(s, outpath):
        with codecs.open(outpath, 'w', encoding='utf8') as fp:
            fp.write(s)
else:
    basestring = str
    unicode = str

    def write_utf8_to(s, outpath):
        with open(outpath, 'w') as fp:
            fp.write(s)


def is_string_type(s):
    return isinstance(s, (basestring, unicode, str))


def ensure_utf8(s):
    # convert from any object to `unicode`
    if not isinstance(s, basestring):
        s = unicode(s)

    if isinstance(s, unicode):
        s = s.encode('utf-8')
    return unicode(s, encoding='utf-8')


u = ensure_utf8



# Used for test purpose
if __name__ == '__main__':
    import os
    import sys
    from easy_karabiner.main import *

    args_num = len(sys.argv) - 1

    if args_num == 0:
        inpath = os.path.join(os.path.dirname(__file__), '..', 'samples/test.py')
    else:
        inpath = sys.argv[1]

    if args_num >= 2:
        outpath = sys.argv[2]
        string = False
    else:
        outpath = None
        string = True

    try:
        main.callback(inpath, outpath, verbose=True, string=string)
    except SystemExit as e:
        print(e.code)



# -*- coding: utf-8 -*-
from __future__ import print_function
from . import exception
from . import def_tag_map
from .basexml import BaseXML


class DefinitionBase(BaseXML):
    """A object represent definition XML node in Karabiner.
    For example, the following XML is a typical definition.

        <appdef>
          <appname>BILIBILI</appname>
          <equal>com.typcn.Bilibili</equal>
        </appdef>
    """
    def __init__(self, name, *tag_vals, **kwargs):
        self.name = name
        self.tag_vals = tag_vals
        self.kwargs = kwargs

    def get_name(self):
        return self.name

    @classmethod
    def get_def_tag_name(cls):
        return '%sdef' % cls.get_class_name().lower()

    @classmethod
    def get_name_tag_name(cls):
        return def_tag_map.get_name_tag_name(cls.get_def_tag_name())

    def get_tag_val_pair(self, val):
        raise exception.NeedOverrideError()

    def get_tag_val_pairs(self, tag_vals):
        return [self.get_tag_val_pair(tag_val) for tag_val in tag_vals]

    @classmethod
    def split_name_and_attrs(cls, tag_name):
        """Because some tag names contain attributes,
        so we need separate it into two parts.
        """
        # transform `key="value"` to `(key, value)`
        def to_pair(s):
            removed_quote = ''.join(ch for ch in s if ch not in ('"', "'"))
            return removed_quote.split('=')

        name_parts = tag_name.split()
        name = name_parts[0]
        raw_attrs = [w for w in name_parts if "=" in w]
        tag_attrs = dict(to_pair(w) for w in raw_attrs)
        return name, tag_attrs

    def to_xml(self):
        xml_tree = self.create_tag(self.get_def_tag_name())
        name_tag = self.create_tag(self.get_name_tag_name(), self.name)
        xml_tree.append(name_tag)

        tag_val_pairs = self.get_tag_val_pairs(self.tag_vals)

        for tag_name, tag_val in tag_val_pairs:
            if len(tag_name) > 0:
                tag_name, tag_attrs = self.split_name_and_attrs(tag_name)
                tag = self.create_tag(tag_name, tag_val, attrib=tag_attrs)
                xml_tree.append(tag)

        return xml_tree

    @property
    def id(self):
        return self.get_def_tag_name(), self.get_name()

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id


class NoNameTagDefinitionBase(DefinitionBase):
    def to_xml(self):
        xml_tree = self.create_tag(self.get_def_tag_name(), self.name)
        return xml_tree


class App(DefinitionBase):
    """
    >>> d = App('BILIBILI', 'com.typcn.Bilibili')
    >>> s = '''
    ...     <appdef>
    ...       <appname>BILIBILI</appname>
    ...       <equal>com.typcn.Bilibili</equal>
    ...     </appdef>
    ...     '''
    >>> util.assert_xml_equal(d, s)
    """

    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        if val[0] == '.':
            return ('suffix', val)
        elif val[-1] == '.':
            return ('prefix', val)
        else:
            return ('equal', val)


class WindowName(DefinitionBase):
    def get_tag_val_pair(self, val):
        return ('regex', val)


class DeviceVendor(DefinitionBase):
    def get_tag_val_pair(self, val):
        return ('vendorid', val)


class DeviceProduct(DefinitionBase):
    def get_tag_val_pair(self, val):
        return ('productid', val)


class InputSource(DefinitionBase):
    @classmethod
    def is_host(cls, val):
        parts = val.split('.')
        return len(parts) >= 3 and all(len(part) for part in parts)

    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        if val[-1] == '.':
            return ('inputsourceid_prefix', val[:-1])
        elif self.is_host(val):
            return ('inputsourceid_equal', val)
        else:
            return ('languagecode', val)


class VKChangeInputSource(InputSource):
    pass


class VKOpenURL(DefinitionBase):
    def get_tag_val_pair(self, val):
        if len(val) == 0:
            return ()
        elif val.startswith('/'):
            return ('url type="file"', val)
        elif val.startswith('#!'):
            if val.startswith('#! '):
                val = val[3:]
            script = self.create_cdata_text(val)
            return ('url type="shell"', script)
        else:
            return ('url', val)

    def to_xml(self):
        xml_tree = super(VKOpenURL, self).to_xml()
        if self.kwargs.get('background', False):
            background_tag = self.create_tag('background')
            xml_tree.append(background_tag)
        return xml_tree


class Replacement(DefinitionBase):
    def get_tag_val_pair(self, val):
        return ('replacementvalue', val)

    def get_tag_val_pairs(self, tag_vals):
        vals = []

        for tag_val in tag_vals:
            val = tag_val.strip()
            if self.is_cdata_text(val):
                val = self.create_cdata_text(self.remove_cdata_mark(val))
            else:
                val = tag_val

            vals.append(val)

        return [self.get_tag_val_pair(', '.join(vals))]


class UIElementRole(NoNameTagDefinitionBase):
    pass


class Modifier(NoNameTagDefinitionBase):
    pass


if __name__ == "__main__":
    import doctest
    from . import util
    doctest.testmod(extraglobs={'util': util})



# -*- coding: utf-8 -*-
from __future__ import print_function


class KeyCombo(object):
    """Convert key list to Karabiner's favorite format

    >>> print(KeyCombo(['ModifierFlag::SHIFT_L', 'ModifierFlag::COMMAND_L']))
    KeyCode::SHIFT_L, ModifierFlag::COMMAND_L

    >>> print(KeyCombo(['ModifierFlag::SHIFT_L', 'ModifierFlag::COMMAND_L', 'KeyCode::K'], has_modifier_none=True))
    KeyCode::K, ModifierFlag::SHIFT_L, ModifierFlag::COMMAND_L, ModifierFlag::NONE
    """

    def __init__(self, keys=None, has_modifier_none=False, keep_first_keycode=False):
        self.has_modifier_none = has_modifier_none
        self.keep_first_keycode = keep_first_keycode
        self.keys = self.parse(keys or [])

    def parse(self, keys):
        """
        :param keys: List[str]
        :return: List[str]
        """
        # we need adjust position of keys, because the first key cannot be modifier key in most case
        keys = self.rearrange_keys(keys)

        if len(keys) > 0:
            if not self.keep_first_keycode and self.is_modifier_key(keys[0]):
                # change the header of first key to `KeyCode`
                keys = self.regularize_first_key(keys)
            if self.has_modifier_none and self.is_modifier_key(keys[-1]):
                keys = self.add_modifier_none(keys)

        return keys

    def rearrange_keys(self, keys):
        tmp = []
        last = 0

        for i in range(len(keys)):
            if not self.is_modifier_key(keys[i]):
                tmp.append(keys[i])
                while last < i:
                    tmp.append(keys[last])
                    last += 1
                last = i + 1

        tmp.extend(keys[last:])
        return tmp

    @classmethod
    def is_modifier_key(cls, key):
        return key.lower().startswith('modifier')

    @classmethod
    def regularize_first_key(cls, keys):
        parts = keys[0].split('::', 1)
        keys[0] = 'KeyCode::' + parts[-1]
        return keys

    # Append 'ModifierFlag::NONE' if you want to change from this key
    # For more information about 'ModifierFlag::NONE', see https://pqrs.org/osx/karabiner/xml.html.en
    @classmethod
    def add_modifier_none(cls, keys):
        keys.append('ModifierFlag::NONE')
        return keys

    def to_str(self):
        return ', '.join(self.keys)

    def __add__(self, another):
        res = KeyCombo()
        res.keys = self.keys + another.keys
        res.has_modifier_none = self.has_modifier_none or another.has_modifier_none
        res.keep_first_keycode = self.keep_first_keycode or another.keep_first_keycode
        return res

    def __str__(self):
        return self.to_str()


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*
from __future__ import print_function
import operator
from functools import reduce
from itertools import groupby

from . import util
from . import query
from . import osxkit
from . import filter
from . import keymap
from . import definition
from . import exception
from . import def_tag_map
from .basexml import BaseXML
from .fucking_string import is_string_type


def define_filters(raw_filters):
    """Found undefined `Filter` in `raw_filters`, and create a `Definition` by it."""
    for raw_filter in raw_filters:
        if raw_filter.startswith('!'):
            val = raw_filter[1:]
        else:
            val = raw_filter

        if not query.is_defined_filter(val):
            FilterCreater.define(val)


def define_keymaps(raw_keymaps):
    """Found undefined `Key` in `raw_keymaps`, and create a `Definition` by it."""
    # raw_keymap: (str, (str), ...)
    for raw_keymap in raw_keymaps:
        try:
            # raw_keycombo: (str)
            for raw_keycombo in raw_keymap[1:]:
                # raw_key: str
                for raw_key in raw_keycombo:
                    if not query.is_defined_key(raw_key):
                        KeymapCreater.define_key(raw_key)
        except exception.UndefinedKeyException as e:
            exception.ExceptionRegister.record_by(raw_keymap, e)


def create_filters(raw_filters):
    """Create `Filter` object from `raw_filters`,
    `Filter` objects with the same class has been merged to one `Filter` object.
    """
    def filter_sort_key(f):
        return f.get_tag_name()

    filter_objs = []

    for raw_filter in raw_filters:
        objs = FilterCreater.create(raw_filter)
        filter_objs.extend(objs)

    grouped = groupby(sorted(filter_objs, key=filter_sort_key), filter_sort_key)
    # merge filters in the same group
    filter_objs = tuple(reduce(operator.add, group) for _, group in grouped)
    return filter_objs


def create_keymaps(raw_keymaps):
    """Create `Keymap` object from `raw_keymaps`."""
    keymap_objs = []

    for raw_keymap in raw_keymaps:
        try:
            keymap_obj = KeymapCreater.create(raw_keymap)
            keymap_objs.append(keymap_obj)
        except exception.InvalidKeymapException as e:
            exception.ExceptionRegister.record_by(raw_keymap, e)

    return keymap_objs


def create_definitions(definitions):
    definition_objs = []

    for name in sorted(definitions.keys()):
        # make sure `vals` is iterable
        if util.is_list_or_tuple(definitions[name]):
            vals = definitions[name]
        else:
            vals = [definitions[name]]

        try:
            tmp = DefinitionCreater.create(name, vals)
            definition_objs.extend(tmp)
        except exception.InvalidDefinition as e:
            k = {name: definitions[name]}
            exception.ExceptionRegister.record(k, e)

    return definition_objs


class FilterCreater(object):
    @classmethod
    def define(cls, val):
        device_ids = osxkit.get_peripheral_info(val)
        # if `val` is a peripheral name
        if device_ids:
            return DefinitionCreater.define_device(val, device_ids)
        else:
            app_info = osxkit.get_app_info(val)
            # if `val` is a application name
            if app_info:
                bundle_id = app_info[1]
                return DefinitionCreater.define_app(val, bundle_id)
            else:
                raise exception.UndefinedFilterException(val)

    @classmethod
    def create(cls, raw_val):
        """Create a list of `Filter` object from `raw_val` string."""
        if raw_val.startswith('!'):
            val = raw_val[1:]
            type = 'not'
        else:
            val = raw_val
            type = 'only'

        class_names = query.query_filter_class_names(val, scope='all')
        # if `val` has been defined
        if class_names:
            filter_classes = [filter.__dict__.get(c) for c in class_names]
            definition_objs = query.DefinitionBucket.get('filter', val)

            # if `val` is user defined
            if definition_objs:
                filter_names = [d.get_name() for d in definition_objs]
            # if `val` is predefined
            else:
                filter_names = [val]

            filter_objs = []
            for i in range(len(filter_classes)):
                filter_class = filter_classes[i]
                filter_name = cls.escape_filter_name(filter_class.__name__, filter_names[i])
                filter_obj = filter_class(filter_name, type=type)
                filter_objs.append(filter_obj)
            return filter_objs
        else:
            raise exception.UndefinedFilterException(raw_val)

    @classmethod
    def escape_filter_name(cls, class_name, name_val):
        if class_name == 'ReplacementFilter':
            name_val = '{{%s}}' % name_val
        elif class_name == 'DeviceVendorFilter':
            name_val = 'DeviceVendor::%s' % name_val
        elif class_name == 'DeviceProductFilter':
            name_val = 'DeviceProduct::%s' % name_val
        elif class_name == 'ModifierFilter':
            name_val = KeymapCreater.get_karabiner_format_key(name_val, key_header='ModifierFlag')
        return name_val


class KeymapCreater(object):
    @classmethod
    def define_key(cls, val):
        app_info = osxkit.get_app_info(val)

        # if `val` is a application name
        if app_info:
            app_path = app_info[0]
            return DefinitionCreater.define_open(app_path, index=val)
        # if `val` is `VKOpenURL` format
        elif DefinitionDetector.is_vkopenurl(val):
            return DefinitionCreater.define_open(val)
        else:
            raise exception.UndefinedKeyException(val)

    @classmethod
    def create(cls, raw_keymap):
        """Create a `Keymap` object from `raw_keymap`."""
        # found the `Keymap` constructor by the command marker
        command = raw_keymap[0].strip('_')
        command = query.get_keymap_alias(command) or command
        keymap_class = keymap.__dict__.get(command)

        new_keycombos = []

        # Translate key string to `Header::Value` format if it has been defined,
        # otherwise, keep it unchanged.
        for raw_keycombo in raw_keymap[1:]:
            new_keycombo = []

            for raw_key in raw_keycombo:
                definition_objs = query.DefinitionBucket.get('key', raw_key)
                if definition_objs:
                    new_keys = [d.get_name() for d in definition_objs]
                else:
                    new_keys = [query.get_key_alias(raw_key) or raw_key]

                for new_key in new_keys:
                    new_key = cls.get_karabiner_format_key(new_key)
                    new_keycombo.append(new_key)

            new_keycombos.append(new_keycombo)

        try:
            if keymap_class:
                return keymap_class(*new_keycombos)
            # if can't found the `Keymap` constructor
            else:
                return keymap.UniversalKeyToKey(command, *new_keycombos)
        except TypeError:
            raise exception.InvalidKeymapException(raw_keymap)

    @classmethod
    def get_karabiner_format_key(cls, key, key_header=None):
        key_parts = key.split('::', 1)

        if len(key_parts) < 2:
            key = query.get_key_alias(key.lower()) or key

            for k in [key, key.upper(), key.lower()]:
                predefined_header = query.KeyHeaderQuery.query(k)
                if predefined_header:
                    return "%s::%s" % (predefined_header, k)

        if key_header:
            return "%s::%s" % (key_header, key_parts[-1])
        else:
            return key


class DefinitionCreater(object):
    @classmethod
    def create(cls, raw_name, vals):
        """Create a list of `Definition` object by found out the relevant constructor intelligently."""
        name_parts = raw_name.split('::', 1)
        # { name : [val] }
        # if no explicit `DefinitionType`, then look at `vals` to figure it out
        if len(name_parts) == 1:
            # { name : val }
            if len(vals) == 1:
                val = vals[0]
                if DefinitionDetector.is_vkopenurl(val):
                    return cls.define('VKOpenURL', raw_name, raw_name, vals)
                elif DefinitionDetector.is_uielementrole(val):
                    return cls.define('UIElementRole', raw_name, raw_name, vals)
                elif DefinitionDetector.is_app(val):
                    return cls.define('App', raw_name, raw_name, vals)
                elif DefinitionDetector.is_replacement(val):
                    return cls.define_replacement(raw_name, vals)
                else:
                    raise exception.InvalidDefinition(raw_name)
            # { name : [val] }
            elif len(vals) > 1:
                if all(is_string_type(val) for val in vals):
                    # { name : [DeviceVendorID, DeviceProductID] }
                    if DefinitionDetector.is_device(vals):
                        return cls.define_device(raw_name, vals)
                    elif DefinitionDetector.is_all_app(vals):
                        return cls.define('App', raw_name, raw_name, vals)

                return cls.define_replacement(raw_name, vals)
        # { DefHeader::DefName : [val] }
        else:
            def_header, def_name = name_parts
            class_name = query.get_def_alias(def_header) or def_header
            definition_objs = cls.define(class_name, raw_name, def_name, vals)
            if definition_objs:
                return definition_objs
            else:
                raise exception.InvalidDefinition(raw_name)

    @classmethod
    def define(cls, class_name, raw_name, def_name, vals, escape_def_name=True):
        def_class = definition.__dict__.get(class_name)
        if def_class:
            if escape_def_name:
                def_name = cls.escape_def_name(def_name, class_name)
            definition_obj = def_class(def_name, *vals)

            def_category = DefinitionDetector.get_definition_caregory(def_class)
            query.DefinitionBucket.put(def_category, raw_name, [definition_obj])
            query.DefinitionBucket.put(def_category, def_name, [definition_obj])
            return [definition_obj]
        else:
            return []

    @classmethod
    def define_replacement(cls, raw_name, vals):
        # if `vals` is `Keymap` format, define and create it
        if all(DefinitionDetector.is_keymap(val) for val in vals):
            from .parse import parse

            block_objs, _ = parse(vals, {})

            keymap_strs = ''
            for block in block_objs:
                keymap_strs += '\n'.join(str(o) for o in block.keymaps)

            after_defined = [BaseXML.create_cdata_text(keymap_strs)]
        else:
            after_defined = []

            for val in vals:
                # if `val` is application name, define it and replace with define name
                app_info = osxkit.get_app_info(val)
                if app_info:
                    bundle_id = app_info[1]
                    DefinitionCreater.define_app(val, bundle_id)
                    after_defined.append(cls.escape_def_name(val, 'App'))
                else:
                    after_defined.append(val)

        return cls.define('Replacement', raw_name, raw_name, after_defined)

    @classmethod
    def define_device(cls, raw_name, vals):
        def_name = cls.escape_def_name(raw_name)
        vdef_name = '%s_VENDOR' % def_name
        pdef_name = '%s_PRODUCT' % def_name

        [vdefinition_obj] = cls.define('DeviceVendor', raw_name, vdef_name, [vals[0]], escape_def_name=False)
        [pdefinition_obj] = cls.define('DeviceProduct', raw_name, pdef_name, [vals[1]], escape_def_name=False)
        definition_objs = (vdefinition_obj, pdefinition_obj)

        query.DefinitionBucket.put('filter', raw_name, definition_objs)
        return definition_objs

    @classmethod
    def define_app(cls, raw_name, bundle_id):
        return cls.define('App', raw_name, raw_name, [bundle_id])

    @classmethod
    def define_open(cls, val, index=None):
        if index:
            return cls.define('VKOpenURL', index, index, [val])
        else:
            def_name = util.get_checksum(val)
            return cls.define('VKOpenURL', val, def_name, [val])

    @classmethod
    def escape_def_name(cls, def_name, class_name=None):
        if class_name == 'VKOpenURL' and not def_name.startswith('KeyCode::VK_OPEN_URL_'):
            def_name = util.escape_string(def_name)
            return 'KeyCode::VK_OPEN_URL_%s' % def_name
        else:
            return util.escape_string(def_name)


class DefinitionDetector(object):
    @classmethod
    def is_device(cls, vals):
        return len(vals) == 2 and all(util.is_hex(val) for val in vals)

    @classmethod
    def is_vkopenurl(cls, val):
        # if `val` is script, application or file, or url
        return any([val.startswith('#! '),
                    val.endswith('.app') or val.startswith('/'),
                    val.startswith('http://') or val.startswith('https://')])

    @classmethod
    def is_uielementrole(cls, val):
        return val.startswith('AX') and val.isalpha()

    @classmethod
    def is_app(cls, val):
        """`appdef` has format like:

             equal:  com.example.application
             prefix: com.example.
             suffix: .example.application
        """
        words = val.split('.')
        if len(words) < 3:
            return False
        else:
            if words[0] == '':
                words.pop(0)
            elif words[-1] == '':
                words.pop()
            return all(w.isalpha() for w in words)

    @classmethod
    def is_all_app(cls, vals):
        return all(cls.is_app(val) for val in vals)

    @classmethod
    def is_replacement(cls, val):
        if BaseXML.is_cdata_text(val):
            return True
        elif cls.is_keymap(val):
            return True
        else:
            return False

    @classmethod
    def is_keymap(cls, val):
        return util.is_list_or_tuple(val)

    @classmethod
    def get_definition_caregory(cls, def_class):
        def_tag_name = def_class.get_def_tag_name()
        if def_tag_map.get_filter_class_name(def_tag_name):
            return 'filter'
        else:
            return 'key'



# -*- coding: utf-8 -*-
from __future__ import print_function
from . import exception
from .basexml import BaseXML
from .keycombo import KeyCombo
from .fucking_string import u


class KeyToKeyBase(BaseXML):
    """A object represent `autogen` XML node in Karabiner.
    For example, the following XML is a typical `autogen`.

        <autogen> __KeyToKey__
            KeyCode::E, ModifierFlag::OPTION_L, ModifierFlag::NONE,
            KeyCode::E, ModifierFlag::COMMAND_L
        </autogen>
    """
    MULTI_KEYS_FMT = u('@begin\n{key}\n@end')
    AUTOGEN_FMT = u(' {type}\n{keys_str}\n')

    def __init__(self, *keycombos, **kwargs):
        self.keys_str = self.parse(*keycombos, **kwargs)

    def parse(self, *keycombos, **kwargs):
        """
        :param keycombos: List[List[str]]
        :param kwargs: Dict
        :return: str
        """
        raise exception.NeedOverrideError()

    def get_type(self):
        return '__%s__' % self.get_class_name()

    def to_xml(self):
        text = self.AUTOGEN_FMT.format(type=self.get_type(),
                                       keys_str=self.keys_str)
        return self.create_tag('autogen', text)


class UniversalKeyToKey(KeyToKeyBase):
    def __init__(self, type, *keycombos, **kwargs):
        self.type = type
        super(UniversalKeyToKey, self).__init__(*keycombos, **kwargs)

    def get_type(self):
        return self.type

    def parse(self, *keycombos, **kwargs):
        keys = [KeyCombo(k, keep_first_keycode=True, has_modifier_none=False) for k in keycombos]
        return ',\n'.join(str(k) for k in keys)


class _KeyToMultiKeys(KeyToKeyBase):
    KEYS_FMT = u('{from_key},\n'
                 '@begin\n'
                 '{to_key}\n'
                 '@end\n'
                 '@begin\n'
                 '{additional_key}\n'
                 '@end')

    def parse(self, from_key, to_key, additional_key, has_modifier_none=True):
        from_key = KeyCombo(from_key, has_modifier_none)
        to_key = KeyCombo(to_key)
        additional_key = KeyCombo(additional_key)

        return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                    to_key=to_key.to_str(),
                                    additional_key=additional_key.to_str())


class _KeyToAdditionalKey(_KeyToMultiKeys):
    def parse(self, from_key, to_key=None, additional_key=None, has_modifier_none=True):
        if to_key is None and additional_key is None:
            raise exception.InvalidKeymapException(from_key)
        elif to_key and additional_key is None:
            additional_key = to_key
            to_key = from_key

        return super(_KeyToAdditionalKey, self).parse(from_key,
                                                      to_key,
                                                      additional_key,
                                                      has_modifier_none)


class _OneKeyEvent(KeyToKeyBase):
    def parse(self, key, has_modifier_none=False):
        return KeyCombo(key, has_modifier_none).to_str()


class _ZeroKeyEvent(KeyToKeyBase):
    def parse(self):
        return ''


class _MultiKeys(UniversalKeyToKey):
    def __init__(self, *keys, **kwargs):
        super(_MultiKeys, self).__init__(self.get_type(), *keys, **kwargs)

    def get_type(self):
        return '__%s__' % self.get_class_name()


class KeyToKey(KeyToKeyBase):
    """
    >>> keymap = KeyToKey(['ModifierFlag::OPTION_L', 'KeyCode::E'],
    ...                   ['ModifierFlag::COMMAND_L', 'KeyCode::E'])
    >>> s = '''
    ...     <autogen> __KeyToKey__
    ...         KeyCode::E, ModifierFlag::OPTION_L, ModifierFlag::NONE,
    ...         KeyCode::E, ModifierFlag::COMMAND_L
    ...     </autogen>
    ...     '''
    >>> util.assert_xml_equal(keymap, s)
    """
    KEYS_FMT = u('{from_key},\n{to_key}')

    def parse(self, from_key, to_key, has_modifier_none=True):
        from_key = KeyCombo(from_key, has_modifier_none=has_modifier_none).to_str()
        to_key = KeyCombo(to_key).to_str()

        return self.KEYS_FMT.format(from_key=from_key, to_key=to_key)


class DropAllKeys(KeyToKeyBase):
    KEYS_FMT = u('{from_key},\n{options}')

    def parse(self, from_modifier, options=None):
        from_key = KeyCombo(from_modifier, keep_first_keycode=True)

        if options is None:
            return from_key.to_str()
        else:
            options = KeyCombo(options, keep_first_keycode=True)
            return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                        options=options.to_str())


class SimultaneousKeyPresses(KeyToKeyBase):
    KEYS_FMT = u('@begin\n'
                 '{from_key}\n'
                 '@end\n'
                 '@begin\n'
                 '{to_key}\n'
                 '@end')

    def parse(self, from_key, to_key):
        from_key = KeyCombo(from_key, False)
        to_key = KeyCombo(to_key, False)

        return self.KEYS_FMT.format(from_key=from_key.to_str(),
                                    to_key=to_key.to_str())


class DoublePressModifier(_KeyToAdditionalKey):
    pass


class HoldingKeyToKey(_KeyToAdditionalKey):
    pass


class KeyOverlaidModifier(_KeyToAdditionalKey):
    pass


class KeyDownUpToKey(_KeyToMultiKeys):
    def parse(self, from_key, immediately_key, interrupted_key=None, has_modifier_none=True):
        return super(KeyDownUpToKey, self).parse(from_key,
                                                 immediately_key,
                                                 interrupted_key or from_key,
                                                 has_modifier_none)


class BlockUntilKeyUp(_OneKeyEvent):
    pass


class DropKeyAfterRemap(_OneKeyEvent):
    pass


class PassThrough(_ZeroKeyEvent):
    pass


class DropPointingRelativeCursorMove(_ZeroKeyEvent):
    pass


# NOTICE: because below `autogen` command not yet documented,
#         so those implements maybe changed in future
class DropScrollWheel(_MultiKeys):
    pass


class FlipPointingRelative(_MultiKeys):
    pass


class FlipScrollWheel(_MultiKeys):
    pass


class IgnoreMultipleSameKeyPress(_MultiKeys):
    pass


class StripModifierFromScrollWheel(_MultiKeys):
    pass


class ScrollWheelToScrollWheel(_MultiKeys):
    pass


class ScrollWheelToKey(_MultiKeys):
    pass


class PointingRelativeToScroll(_MultiKeys):
    pass


class PointingRelativeToKey(_MultiKeys):
    pass


class ForceNumLockOn(_MultiKeys):
    pass


class ShowStatusMessage(KeyToKeyBase):
    def parse(self, *keys, **kwargs):
        return ' '.join(keys)


class SetKeyboardType(_OneKeyEvent):
    pass


if __name__ == "__main__":
    import doctest
    from . import util
    doctest.testmod(extraglobs={'util': util})



# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import glob
from itertools import chain
from . import alias
from . import config
from . import exception
from . import definition
from . import def_tag_map
from .basexml import BaseXML


def is_defined_filter(val):
    return query_filter_class_names(val, scope='all')


def is_defined_key(val):
    if get_key_alias(val.lower()):
        return True
    else:
        for k in [val, val.upper(), val.lower()]:
            if KeyHeaderQuery.query(k):
                return True

    return DefinitionBucket.has('key', val)


def is_predefined_modifier(key):
    key = get_key_alias(key.lower()) or key

    for k in [key, key.upper(), key.lower()]:
        if KeyHeaderQuery.query(k) == 'ModifierFlag':
            return True

    parts = key.split('::', 1)
    return len(parts) == 2 and parts[0] == 'ModifierFlag'


def query_filter_class_names(value, scope='all'):
    """Get `Filter` class name by `value` in `scope`."""
    definitions = DefinitionBucket.get('filter', value)
    def_tag_name = DefinitionTypeQuery.query(value)
    filter_class_names = []

    if scope in ('all', 'user') and definitions:
        for d in definitions:
            filter_class_name = def_tag_map.get_filter_class_name(d.get_def_tag_name())
            if filter_class_name:
                filter_class_names.append(filter_class_name)
            else:
                return []
    elif scope in ('all', 'predefined'):
        if not def_tag_name:
            if is_predefined_modifier(value):
                def_tag_name = definition.Modifier.get_def_tag_name()
            else:
                return []

        filter_class_name = def_tag_map.get_filter_class_name(def_tag_name)
        if filter_class_name:
            filter_class_names.append(filter_class_name)

    return filter_class_names


def get_def_alias(value):
    return alias.get_alias('DEF_ALIAS', value)


def get_key_alias(value):
    return alias.get_alias('KEY_ALIAS', value)


def get_keymap_alias(value):
    return alias.get_alias('KEYMAP_ALIAS', value)


def update_aliases(aliases_table):
    alias.update_aliases(aliases_table)


class BaseTypeQuery(object):
    """This class is used to get corresponding type of `value` by searching in XML files.
    Because there is a possibility that multiple types associated with the same value,
    for avoid this problem, we specify the order of types and return the first occurrence as the query result.
    """
    DATA_DIR = config.get_karabiner_resources_dir()

    def __init__(self):
        self.data = {}

    @classmethod
    def get_instance(cls):
        if hasattr(cls, '_instance'):
            return cls._instance
        else:
            cls._instance = cls()
            cls._instance.load_data()
            return cls._instance

    @classmethod
    def query(cls, value):
        self = cls.get_instance()
        for type in self.orders:
            if self.is_in(type, value):
                return type
        return None

    def is_in(self, type, value):
        return type if value in self.data.get(type, []) else None

    @property
    def orders(self):
        raise exception.NeedOverrideError()

    def load_data(self):
        raise exception.NeedOverrideError()


class KeyHeaderQuery(BaseTypeQuery):
    RESOURCE_FILE = 'symbol_map.xml'

    @property
    def orders(self):
        return [
            'ModifierFlag',
            'ConsumerKeyCode',
            'KeyCode',
            'Option',
            'InputSource',
            'PointingButton',
            'DeviceProduct',
            'DeviceVendor',
            'KeyboardType',
        ]

    def load_data(self):
        filepath = os.path.join(self.DATA_DIR, self.RESOURCE_FILE)

        xml_tree = BaseXML.parse(filepath)
        for symbol_map in xml_tree:
            type = symbol_map.get('type')
            value = symbol_map.get('name')
            self.data.setdefault(type, set()).add(value)


class DefinitionTypeQuery(BaseTypeQuery):
    @property
    def orders(self):
        return [
            'appdef',
            'replacementdef',
            'modifierdef',
            'devicevendordef',
            'deviceproductdef',
            'uielementroledef',
            'windownamedef',
            'vkopenurldef',
            'inputsourcedef',
            'vkchangeinputsourcedef',
        ]

    @classmethod
    def query(cls, value):
        result = super(DefinitionTypeQuery, cls).query(value)
        if result:
            return result
        elif KeyHeaderQuery.query(value) == 'ModifierFlag':
            return 'modifierdef'

    def load_data(self):
        for filepath in glob.glob(os.path.join(self.DATA_DIR, '*def.xml')):
            type, _ = os.path.splitext(os.path.basename(filepath))
            self.data[type] = self.get_data(type, filepath)

    def get_data(self, type, filepath):
        name_tag = def_tag_map.get_name_tag_name(type)
        xml_tree = BaseXML.parse(filepath)

        if name_tag == '':
            tags = xml_tree.findall(type)
        else:
            tags = xml_tree.findall('%s/%s' % (type, name_tag))

        return set(tag.text for tag in tags)


class DefinitionBucket(object):
    """This class is used to store global `Definition` objects,
    so we can create a `Definition` object from anywhere,
    and found it by the original value used to define.
    """
    def __init__(self):
        self.buckets = {
            'filter': {},
            'key': {},
        }

    @classmethod
    def get_instance(cls, reset=False):
        if not hasattr(cls, '_instance') or reset:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def get_all_definitions(cls):
        list_of_defos = [d.values() for d in cls.get_instance().buckets.values()]
        defos = set(chain.from_iterable(chain.from_iterable(list_of_defos)))
        return sorted(defos, key=lambda f: (f.get_def_tag_name(), f.get_name()))

    @classmethod
    def put(cls, category, name, definitions):
        cls.get_instance().buckets[category][name] = definitions

    @classmethod
    def get(cls, category, name):
        return cls.get_instance().buckets[category].get(name)

    @classmethod
    def has(cls, category, name):
        return name in cls.get_instance().buckets[category]

    @classmethod
    def clear(cls):
        cls.get_instance(reset=True)



# -*- coding: utf-8 -*-
from __future__ import print_function
from . import util
from . import config


def _apply_if_is_alias(vars, func):
    result = {}

    for alias_name in vars.keys():
        alias = vars[alias_name]
        if hasattr(alias, '__iter__') \
                and alias_name.endswith('ALIAS') \
                and not alias_name.startswith('_'):
            result[alias_name] = func(vars, alias_name)

    return result


def _get_aliases():
    if not hasattr(_get_aliases, 'aliases'):
        aliases = util.read_python_file(config.get_data_path('alias.data'))
        _get_aliases.aliases = _apply_if_is_alias(aliases, lambda d, k: d[k])

    return _get_aliases.aliases


# alias is case-insensitive
def get_alias(alias_name, value):
    def get(aname, avalue):
        return _get_aliases()[aname].get(avalue.lower())

    if alias_name == 'KEY_ALIAS':
        return get('KEY_ALIAS', value) or get('MODIFIER_ALIAS', value)
    else:
        return get(alias_name, value)


def update_alias(alias_name, new_aliases):
    _get_aliases().setdefault(alias_name, {}).update(new_aliases)


def update_aliases(aliases_table):
    _apply_if_is_alias(aliases_table, lambda d, k: update_alias(k, d[k]))



# -*- coding: utf-8 -*-
"""A tool to generate key remap configuration for Karabiner

Usage:
    easy_karabiner [-evrl] [SOURCE] [TARGET | --string]
    easy_karabiner [--help | --version]
"""
from __future__ import print_function
import os
import sys
import lxml
import click
from subprocess import call
from . import __version__
from . import util
from . import query
from . import config
from . import exception
from .osxkit import get_all_peripheral_info
from .basexml import BaseXML
from .generator import Generator
from .fucking_string import *


@click.command()
@click.help_option('--help', '-h')
@click.version_option(__version__, '--version', '-v', message='%(version)s')
@click.argument('inpath', default=config.get_default_config_path(), type=click.Path())
@click.argument('outpath', default=config.get_default_output_path(), type=click.Path())
@click.option('--verbose', '-V', help='Print more text.', is_flag=True)
@click.option('--string', '-s', help='Output as string.', is_flag=True)
@click.option('--reload', '-r', help='Reload Karabiner.', is_flag=True)
@click.option('--no-reload', help='Opposite of --reload', is_flag=True)
@click.option('--edit', '-e', help='Edit default config file.', is_flag=True)
@click.option('--list-peripherals', '-l', help='List name of all peripherals.', is_flag=True)
def main(inpath, outpath, **options):
    """
    \b
    $ easy_karabiner
    $ easy_karabiner input.py output.xml
    $ easy_karabiner input.py --string
    """
    config.set(options)

    if config.get('help') or config.get('version'):
        return
    elif config.get('edit'):
        edit_config_file()
    elif config.get('list_peripherals'):
        list_peripherals()
    elif config.get('reload'):
        reload_karabiner()
    else:
        try:
            configs = read_config_file(inpath)
            xml_str = gen_config(configs)

            if config.get('string'):
                print_message(xml_str)
            else:
                try:
                    if not is_generated_by_easy_karabiner(outpath):
                        backup_file(outpath)
                except IOError:
                    pass
                finally:
                    write_generated_xml(outpath, xml_str)
                    if is_need_reload(config.get('reload'), config.get('no_reload'), outpath):
                        reload_karabiner()

            show_config_warnings()
        except exception.ConfigError as e:
            print_error(e)
            sys.exit(1)
        except IOError as e:
            print_error("%s not exist" % e.filename)
            sys.exit(1)
        except Exception as e:
            print_error(e)
            sys.exit(1)

    sys.exit(0)


def read_config_file(config_path):
    if config.get('verbose'):
        print_info('Execute "%s"' % config_path)
    return util.read_python_file(config_path)


def write_generated_xml(outpath, content):
    if config.get('verbose'):
        print_info('Write XML to "%s"' % outpath)
    write_utf8_to(content, outpath)


def edit_config_file():
    click.edit(filename=config.get_default_config_path())


def is_need_reload(reload, no_reload, outpath):
    if no_reload:
        return False
    else:
        return reload or (outpath == config.get_default_output_path())


def reload_karabiner():
    NOTIFICATION_MSG = "Enabled generated configuration"
    NOTIFICATION_CMD = ('/usr/bin/osascript', '-e',
                        'display notification "%s" with title "Karabiner Reloaded"' % NOTIFICATION_MSG)
    KARABINER_CMD = config.get_karabiner_bin('karabiner')

    if config.get('verbose'):
        print_info("Reload Karabiner config")
    call([KARABINER_CMD, 'enable', 'private.easy_karabiner'])
    call([KARABINER_CMD, 'reloadxml'])
    call(NOTIFICATION_CMD)


def list_peripherals():
    peripheral_names = [ensure_utf8(name) for name in get_all_peripheral_info().keys()]
    for name in sorted(peripheral_names):
        print_message(name)


def gen_config(configs):
    maps = configs.get('MAPS')
    definitions = configs.get('DEFINITIONS')
    query.update_aliases(configs)
    if config.get('verbose'):
        print_info("Generate XML configuration")
    return Generator(maps, definitions).to_str()


def is_generated_by_easy_karabiner(filepath):
    try:
        tag = BaseXML.parse(filepath).find('Easy-Karabiner')
        return tag is not None
    except lxml.etree.XMLSyntaxError:
        return False


def backup_file(filepath, new_path=None):
    with open(filepath, 'rb') as fp:
        if new_path is None:
            # private.xml -> private.941f123.xml
            checksum = util.get_checksum(fp.read())
            parts = os.path.basename(filepath).split('.')
            parts.insert(-1, checksum)
            new_name = '.'.join(parts)

            if config.get('verbose'):
                print_info("Backup original XML config file")
            new_path = os.path.join(os.path.dirname(filepath), new_name)

        os.rename(filepath, new_path)
        return new_path


def show_config_warnings():
    records = exception.ExceptionRegister.get_all_records()
    for raw_data, e in records:
        exception_class = type(e)

        if exception_class == exception.UndefinedFilterException:
            msg = 'Undefined filter'
        elif exception_class == exception.UndefinedKeyException:
            msg = 'Undefined key'
        elif exception_class == exception.InvalidDefinition:
            msg = 'Invalid definition'
        elif exception_class == exception.InvalidKeymapException:
            msg = 'Invalid keymap'
        else:
            msg = exception_class.__name__

        if len(e.args) == 0:
            print_warning('%s: %s' % (msg, raw_data))
        elif len(e.args) == 1:
            print_warning('%s: `%s` in %s' % (msg, e.args[0], raw_data))
        else:
            print_warning('%s: %s in %s' % (msg, e.args, raw_data))


def print_message(msg, color=None, err=False):
    """Seems `click.echo` has fixed the problem of UnicodeDecodeError when redirecting (See
    https://stackoverflow.com/questions/4545661/unicodedecodeerror-when-redirecting-to-file
    for detail). As a result, the below code used to solve the problem is conflict with `click.echo`.
    To avoid the problem, you should always use `print` with below code or `click.echo` in `__main__.py`

        if sys.version_info[0] == 2:
            sys.stdout = codecs.getwriter("utf-8")(sys.stdout)
        else:
            sys.stdout = codecs.getwriter("utf-8")(sys.stdout.detach())
    """
    click.secho(msg, fg=color, err=err)


def print_error(msg):
    print_message(msg, color='red', err=True)


def print_warning(msg):
    print_message(msg, color='yellow', err=True)


def print_info(msg):
    print_message(msg, color='green')



# -*- coding: utf-8 -*-
from __future__ import print_function
from .fucking_string import ensure_utf8


class NeedOverrideError(NotImplementedError):
    def __init__(self, errmsg='You need override this method'):
        super(NeedOverrideError, self).__init__(self, errmsg)


class ConfigError(Exception):
    pass


class ConfigWarning(Exception):
    pass


class InvalidDefinition(ConfigWarning):
    pass


class InvalidKeymapException(ConfigWarning):
    pass


class UndefinedFilterException(ConfigWarning):
    pass


class UndefinedKeyException(ConfigWarning):
    pass


class ExceptionRegister(object):
    """This class is used to record the occurrence of any exceptions,
    so we can provide a user friendly error message.
    """
    error_table = {}
    raw_maps_table = {}

    @classmethod
    def put(cls, k, raw_map):
        k = ensure_utf8(k)
        if k not in cls.raw_maps_table:
            cls.raw_maps_table[k] = raw_map

    @classmethod
    def record_by(cls, k, exception):
        k = ensure_utf8(k)
        raw_map = cls.raw_maps_table[k]
        cls.record(raw_map, exception)

    @classmethod
    def record(cls, k, exception):
        k = ensure_utf8(k)
        if k not in cls.error_table:
            cls.error_table[k] = exception

    @classmethod
    def get_all_records(cls):
        return cls.error_table.items()



# -*- coding: utf-8 -*-
import os

KARABINER_APP_PATH = '/Applications/Karabiner.app'
DEFAULT_CONFIG_PATH = '~/.easy_karabiner.py'
DEFAULT_OUTPUT_PATH = '~/Library/Application Support/Karabiner/private.xml'

command_options = {}


def set(options):
    command_options.update(options)


def get(key):
    return command_options.get(key)


def get_default_config_path():
    return os.path.expanduser(DEFAULT_CONFIG_PATH)


def get_default_output_path():
    return os.path.expanduser(DEFAULT_OUTPUT_PATH)


def get_data_dir():
    return os.path.join(os.path.dirname(__file__), 'data')


def get_data_path(filename):
    return os.path.join(get_data_dir(), filename)


def get_karabiner_app_path():
    return KARABINER_APP_PATH


def _get_path_relate_to_karabiner(relative_path):
    return os.path.join(get_karabiner_app_path(), relative_path)


def get_karabiner_bin_dir():
    return _get_path_relate_to_karabiner('Contents/Library/bin')


def get_karabiner_resources_dir():
    return _get_path_relate_to_karabiner('Contents/Resources')


def get_karabiner_bin(filename):
    return os.path.join(get_karabiner_bin_dir(), filename)



# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import OrderedDict
from operator import add
from functools import reduce

from . import util
from . import query
from . import osxkit
from . import factory
from . import exception
from .basexml import BaseXML


def parse(maps, definitions):
    """Parse and convert user Easy-Karabiner config to XML objects."""
    definitions = util.encode_with_utf8(definitions)
    maps = util.encode_with_utf8(maps)

    factory.create_definitions(definitions)
    filters_keymaps_table = organize_maps(maps)
    block_objs = create_blocks(filters_keymaps_table)

    definition_objs = query.DefinitionBucket.get_all_definitions()
    return block_objs, definition_objs


def organize_maps(maps):
    """Convert `maps` to a `OrderedDict` with the consistent type.
    The items in `filters_keymaps_table` is ordered by the first present of `filters`.
    """
    filters_keymaps_table = OrderedDict()

    for raw_map in maps:
        if util.is_list_or_tuple(raw_map) and len(raw_map) > 0:
            try:
                raw_keymap, raw_filters = separate_keymap_filters(raw_map)
                filters_keymaps_table.setdefault(raw_filters, []).append(raw_keymap)
            except exception.ConfigWarning as e:
                exception.ExceptionRegister.record(raw_map, e)
            else:
                if raw_filters:
                    exception.ExceptionRegister.put(raw_filters, raw_map)
                if raw_keymap:
                    exception.ExceptionRegister.put(raw_keymap, raw_map)
        else:
            raise exception.ConfigError('Map must be a list: %s' % raw_map.__repr__())

    return filters_keymaps_table


def separate_keymap_filters(raw_map):
    """Convert `raw_map` to a `tuple` with consistent type."""
    # if last element in `raw_map` is a filter
    if util.is_list_or_tuple(raw_map[-1]):
        raw_filters = tuple(raw_map[-1])
        raw_map = raw_map[:-1]
    else:
        raw_filters = tuple()

    if len(raw_map) == 0 or len(raw_map[0]) == 0:
        raise exception.ConfigWarning("Cannot found keymap")
    else:
        # if first part in `raw_map` is command marker
        if len(raw_map[0]) > 2 and raw_map[0].startswith('_') and raw_map[0].endswith('_'):
            command = raw_map[0]
            raw_keycombos = raw_map[1:]
        else:
            command = '__KeyToKey__'
            raw_keycombos = raw_map

        # ((key1, key2, ...), (key1, key2, ...))
        keycombos = []
        for keycombo_str in raw_keycombos:
            if factory.DefinitionDetector.is_vkopenurl(keycombo_str) or osxkit.get_app_info(keycombo_str):
                keycombos.append((keycombo_str,))
            else:
                keycombos.append(tuple(util.split_ignore_quote(keycombo_str)))

        raw_keymap = (command,) + tuple(keycombos)
        return raw_keymap, raw_filters


def create_blocks(filters_keymaps_table):
    tmp = OrderedDict()

    for raw_filters in filters_keymaps_table:
        raw_keymaps = filters_keymaps_table[raw_filters]

        try:
            factory.define_filters(raw_filters)
            factory.define_keymaps(raw_keymaps)

            filter_objs = factory.create_filters(raw_filters)
            keymap_objs = factory.create_keymaps(raw_keymaps)
            tmp.setdefault(filter_objs, []).append(keymap_objs)
        except exception.UndefinedFilterException as e:
            exception.ExceptionRegister.record_by(raw_filters, e)

    blocks = []
    for filter_objs in tmp:
        keymap_objs = reduce(add, tmp[filter_objs])
        block = Block(keymap_objs, filter_objs)
        blocks.append(block)
    return blocks


class Block(BaseXML):
    """Block is a kind of XML node similar to `item` in Karabiner.
    For example, the following XML is a typical `block`.

        <block>
            <only>VIRTUALMACHINE</only>
            <autogen> __KeyToKey__
                KeyCode::F, ModifierFlag::CONTROL_L, ModifierFlag::COMMAND_L, ModifierFlag::NONE,
                KeyCode::F, ModifierFlag::COMMAND_L
            </autogen>
        </block>
    """
    def __init__(self, keymaps, filters=None):
        self.keymaps = keymaps
        self.filters = filters or tuple()

    def to_xml(self):
        xml_tree = self.create_tag('block')

        for f in self.filters:
            xml_tree.append(f.to_xml())
        for k in self.keymaps:
            xml_tree.append(k.to_xml())

        return xml_tree



# coding: utf-8
from __future__ import print_function
import os
import subprocess
from . import util
from .fucking_string import ensure_utf8

__all__ = ['get_app_info', 'get_all_app_info',
           'get_peripheral_info', 'get_all_peripheral_info']


def call(cmd, **kwargs):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, **kwargs)
    output = proc.stdout.read().decode('utf-8')
    return output


def contains_any_keywords(s, *keywords):
    words = set(w.lower() for w in s.split())
    keywords = set(w.lower() for w in keywords)
    return not words.isdisjoint(keywords)


def get_name_without_suffix(path):
    return os.path.splitext(os.path.basename(path))[0]


def get_all_app_info():
    """Return dict contains applications information. Item in the dict has format like:

        'basename / display_name' : (absolute_path, bundle_identifier)

    NOTICE: `basename` or `display_name` already removed '.app' suffix
    """
    # list absolute path of all applications
    LIST_APP_PATH_CMD = ('mdfind', '-0', 'kMDItemContentType==com.apple.application-bundle')
    # get Bundle-Identifier and Display-Name of applications by absolute path
    LIST_APP_INFO_CMD = ('mdls', '-raw',
                         '-name', 'kMDItemCFBundleIdentifier',
                         '-name', 'kMDItemDisplayName')

    raw_paths = call(LIST_APP_PATH_CMD).split('\x00')
    app_paths = [p for p in raw_paths if p.strip()]

    raw_infos = call(LIST_APP_INFO_CMD + tuple(app_paths))
    info_lines = raw_infos.split('\x00')

    app_ids = info_lines[0::2]
    app_infos = tuple(zip(app_paths, app_ids))
    # display name not always equal to basename,
    # because system language and application name maybe not English
    display_names = tuple(info_lines[1::2])
    basenames = tuple(get_name_without_suffix(p) for p in app_paths)

    name_map = tuple(zip(display_names + basenames, app_infos + app_infos))
    return dict(name_map)


def get_app_info(name):
    if not hasattr(get_app_info, 'infos'):
        get_app_info.infos = get_all_app_info()

    if name.endswith('.app'):
        name = name[:-4]
    name = ensure_utf8(name)
    return get_app_info.infos.get(name)


def lines2tree(lines):
    def do_lines2tree(root, lines, i=0, level=-1):
        while i < len(lines):
            line = lines[i]
            indent_level = len(line) - len(line.lstrip())

            if indent_level > level:
                key, val = line.split(':', 1)
                key = key.strip()
                val = val.strip()

                if line.endswith(':'):
                    root[key] = {}
                    i = do_lines2tree(root[key], lines, i + 1, indent_level)
                else:
                    root[key] = val
                    i += 1
            else:
                break
        return i

    tree = {}
    do_lines2tree(tree, lines)
    return tree


def get_devices(item, conditions):
    devices = {}
    for k in item:
        if isinstance(item[k], dict) and any(fn(k) for fn in conditions):
            devices[k] = item[k]
    return devices


def get_display_name(key, brand, product_id):
    id_num = product_id[-4:]

    if key == 'Apple Internal Keyboard / Trackpad':
        name = 'built-in keyboard and trackpad'
    elif contains_any_keywords(key, 'Device'):
        name = '%s_%s' % (brand, id_num)
    else:
        name = '%s_%s' % (key, id_num)

    return util.escape_string(name)


def get_all_peripheral_info():
    """Return dict contains devices information. Item in the dict has format like:

        'display_name' : (vendor_id, product_id)
    """
    # list information about USB and Bluetooth devices
    CMD = ('system_profiler', '-detailLevel', 'mini', '-timeout', '1',
           'SPUSBDataType', 'SPBluetoothDataType')

    output = call(CMD)
    lines = [line for line in output.split('\n') if line.strip()]
    info_tree = lines2tree(lines)

    # Bluetooth Devices
    item = [v for k, v in info_tree['Bluetooth'].items() if k.startswith('Devices')][0]
    conditions = [
        lambda k: k == 'Apple Internal Keyboard / Trackpad',
        lambda k: item[k].get('Major Type') == 'Peripheral',
    ]
    devices = get_devices(item, conditions)

    # USB Devices
    for item in info_tree['USB'].values():
        conditions = [
            lambda k: contains_any_keywords(k, 'Keyboard', 'Trackpad', 'Mouse'),
            lambda k: item[k].get('Built-In') != 'Yes',
        ]
        devices.update(get_devices(item, conditions))

    # only keep (Vendor ID, Product ID) properties
    for k in sorted(devices.keys()):
        device = devices.pop(k)
        if 'Product ID' in device and 'Vendor ID' in device:
            product_id = device['Product ID']
            vendor_parts = device['Vendor ID'].split(' ', 1)
            vendor_id = vendor_parts[0]
            brand = device.get('Manufacturer', vendor_parts[-1].strip()[1:-1])
            # construct display name from properties
            name = ensure_utf8(get_display_name(k, brand, product_id))

            devices[name] = (vendor_id, product_id)

    return devices


def get_peripheral_info(name):
    if not hasattr(get_peripheral_info, 'infos'):
        get_peripheral_info.infos = get_all_peripheral_info()

    name = ensure_utf8(name)
    return get_peripheral_info.infos.get(name)


# used for test purpose
if __name__ == '__main__':
    app_names = ['Finder', 'Xee³',
                 '虾米音乐', '坚果云', '地图',
                 'xiami', 'Nutstore', 'Maps']

    peripheral_names = ['“fz”的键盘_0255', 'Lenovo_USB_Optical_Mouse_6019',
                        'built_in_keyboard_and_trackpad', 'CHERRY_GmbH_0011']

    print(' Application Info '.center(80, '-'))
    for app_name in app_names:
        print('%s    \t%s' % (app_name, get_app_info(app_name)))

    print(' Peripheral Info '.center(80, '-'))
    for peripheral_name in peripheral_names:
        print('%-42s\t%s' % (peripheral_name, get_peripheral_info(peripheral_name)))



# -*- coding: utf-8 -*-
from __future__ import print_function
from . import __version__
from . import parse
from .basexml import BaseXML


class Generator(BaseXML):
    """Construct Karabiner favorite XML tree

    >>> g = Generator()
    >>> s = '''
    ...     <root>
    ...       <Easy-Karabiner>{version}</Easy-Karabiner>
    ...       <item>
    ...         <name>Easy-Karabiner</name>
    ...         <item>
    ...           <name>Enable</name>
    ...           <identifier>private.easy_karabiner</identifier>
    ...         </item>
    ...       </item>
    ...     </root>
    ...     '''.format(version=__version__)
    >>> util.assert_xml_equal(g, s)
    """

    ITEM_IDENTIFIER = 'private.easy_karabiner'

    def __init__(self, maps=None, definitions=None):
        self.maps = maps or []
        self.definitions = definitions or {}

    def init_xml_tree(self, xml_root):
        """
            <Easy-Karabiner>{version}</Easy-Karabiner>
            <item>
                <name>Easy-Karabiner</name>
                <!-- REMAIN NODES HERE -->
            </item>
        """
        version_tag = BaseXML.create_tag('Easy-Karabiner', __version__)
        item_tag = BaseXML.create_tag('item')
        item_tag.append(BaseXML.create_tag('name', 'Easy-Karabiner'))
        xml_root.append(version_tag)
        xml_root.append(item_tag)
        return item_tag

    def init_subitem_tag(self, item_tag):
        """
        <item>
            <name>Easy-Karabiner</name>
            <item>
                <name>Enable</name>
                <identifier>private.easy_karabiner</identifier>
                <!-- REMAIN NODES HERE -->
            </item>
        </item>
        """
        subitem_tag = BaseXML.create_tag('item')
        name_tag = BaseXML.create_tag('name', 'Enable')
        identifier_tag = BaseXML.create_tag('identifier', self.ITEM_IDENTIFIER)
        item_tag.append(subitem_tag)
        subitem_tag.append(name_tag)
        subitem_tag.append(identifier_tag)
        return subitem_tag

    def to_xml(self):
        blocks, definitions = parse.parse(self.maps, self.definitions)

        # construct XML tree
        xml_tree = BaseXML.create_tag('root')

        item_tag = self.init_xml_tree(xml_tree)
        for definition in definitions:
            item_tag.append(definition.to_xml())

        subitem_tag = self.init_subitem_tag(item_tag)
        for block in blocks:
            subitem_tag.append(block.to_xml())

        return xml_tree


if __name__ == '__main__':
    import doctest
    from . import util
    doctest.testmod(extraglobs={
        '__version__': __version__,
        'util': util,
    })



# -*- coding: utf-8 -*-
from __future__ import print_function
from . import util
from . import config


def _get_def_tag_maps():
    if not hasattr(_get_def_tag_maps, 'maps'):
        maps = util.read_python_file(config.get_data_path('def_tag_map.data'))['DEF_TAG_MAP']
        _get_def_tag_maps.maps = maps
    return _get_def_tag_maps.maps


def _get(def_tag):
    return _get_def_tag_maps().get(def_tag, [None, None])


def get_name_tag_name(def_tag):
    return _get(def_tag)[0]


def get_filter_class_name(def_tag):
    return _get(def_tag)[1]



# -*- coding: utf-8 -*-
from __future__ import print_function
import shlex
from hashlib import sha1
from .basexml import BaseXML
from .fucking_string import ensure_utf8, is_string_type


def read_python_file(pypath):
    vars = {}
    with open(pypath, 'rb') as fp:
        exec(compile(fp.read(), pypath, 'exec'), {}, vars)
    return vars


def get_checksum(s):
    return sha1(ensure_utf8(s).encode('utf-8')).hexdigest()[:7]


def escape_string(s):
    """Keep char unchanged if char is number or letter or unicode"""
    chs = []
    for ch in s:
        ch = ch if ord(ch) > 128 or ch.isalnum() else ' '
        chs.append(ch)

    # remove multiple whitespaces and replace whitespace with '_'
    return '_'.join(''.join(chs).split())


def encode_with_utf8(o):
    """Encode object `o` with UTF-8 recursively"""
    if is_string_type(o):
        return ensure_utf8(o)

    if is_list_or_tuple(o):
        return type(o)([encode_with_utf8(item) for item in o])
    elif isinstance(o, dict):
        for k in o.keys():
            o[encode_with_utf8(k)] = encode_with_utf8(o.pop(k))
        return o
    else:
        raise TypeError('Cannot encode %s with UTF-8' % o.__repr__())


def is_hex(s):
    try:
        _ = int(s, 16)
        return True
    except ValueError:
        return False


def is_list_or_tuple(obj):
    return isinstance(obj, (list, tuple))


def split_ignore_quote(s):
    return shlex.split(s)


def remove_all_space(s):
    return ''.join(s.split())


def is_xml_element_equal(node1, node2):
    if len(node1) != len(node2):
        return False
    if node1.tag != node2.tag:
        return False
    if node1.attrib != node2.attrib:
        return False

    text1 = '' if node1.text is None else remove_all_space(node1.text)
    text2 = '' if node2.text is None else remove_all_space(node2.text)
    return text1 == text2


def is_xml_tree_equal(tree1, tree2, ignore_tags=tuple()):
    if tree1.tag == tree2.tag and tree1.tag in ignore_tags:
        return True
    elif is_xml_element_equal(tree1, tree2):
        elems1 = list(tree1)
        elems2 = list(tree2)

        for i in range(len(elems1)):
            if not is_xml_tree_equal(elems1[i], elems2[i], ignore_tags=ignore_tags):
                return False
        return True
    else:
        return False


def assert_xml_equal(xml_tree1, xml_tree2, ignore_tags=tuple()):
    if isinstance(xml_tree1, BaseXML):
        xml_tree1 = xml_tree1.__str__()
    if isinstance(xml_tree2, BaseXML):
        xml_tree2 = xml_tree2.__str__()

    nospaces1 = ''.join(xml_tree1.split())
    nospaces2 = ''.join(xml_tree2.split())
    xml_tree1 = BaseXML.parse_string(xml_tree1)
    xml_tree2 = BaseXML.parse_string(xml_tree2)

    if nospaces1 != nospaces2:
        assert(is_xml_tree_equal(xml_tree1, xml_tree2, ignore_tags=ignore_tags))



from setuptools import setup, find_packages
from easy_karabiner import __version__

setup(
    name='easy_karabiner',
    version=__version__,
    description='A tool to simplify key-remapping configuration for Karabiner',
    author='loggerhead',
    author_email='i@loggerhead.me',
    url='https://github.com/loggerhead/Easy-Karabiner',
    keywords=('Karabiner', 'configer', 'remap', 'key'),
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=[
        'lxml >= 3.0.0, < 4.0.0',
        'click >= 6.0.0, < 7.0.0',
    ],
    entry_points='''
        [console_scripts]
        easy_karabiner=easy_karabiner.main:main
    ''',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
    ]
)



# -*- coding: utf-8 -*-
from __future__ import print_function
import lxml.etree as etree
import xml.dom.minidom as minidom
import xml.sax.saxutils as saxutils
from . import exception
from .fucking_string import ensure_utf8


class BaseXML(object):
    xml_parser = etree.XMLParser(strip_cdata=False)

    @classmethod
    def unescape(cls, s):
        return saxutils.unescape(s, {
                "&quot;": '"',
                "&apos;": "'",
            })

    @classmethod
    def parse(cls, filepath):
        return etree.parse(filepath).getroot()

    @classmethod
    def parse_string(cls, xml_str):
        return etree.fromstring(xml_str, cls.xml_parser)

    @classmethod
    def get_class_name(cls):
        return cls.__name__

    @classmethod
    def is_cdata_text(cls, text):
        return text.startswith('<![CDATA[') and text.endswith(']]>')

    @classmethod
    def remove_cdata_mark(cls, text):
        return text[len('<![CDATA['):-len(']]>')]

    @classmethod
    def create_cdata_text(cls, text):
        # do NOT use `etree.CDATA`
        return '<![CDATA[%s]]>' % text

    @classmethod
    def assign_text_attribute(cls, etree_element, text):
        if text is not None:
            etree_element.text = ensure_utf8(text)
        else:
            etree_element.text = text

    @classmethod
    def create_tag(cls, name, text=None, **kwargs):
        et = etree.Element(name, **kwargs)
        cls.assign_text_attribute(et, text)
        return et

    @classmethod
    def pretty_text(cls, elem, indent="  ", level=0):
        """WARNING: This method would change the construct of XML tree"""
        i = "\n" + level * indent

        if len(elem) == 0:
            if elem.text is not None:
                lines = elem.text.split('\n')
                if len(lines) > 1:
                    if not lines[0].startswith(' '):
                        lines[0] = (i + indent) + lines[0]
                    if lines[-1].strip() == '':
                        lines.pop()
                    elem.text = (i + indent).join(lines) + i
        else:
            for subelem in elem:
                BaseXML.pretty_text(subelem, indent, level + 1)

        return elem

    @classmethod
    def to_format_str(cls, xml_tree, pretty_text=True):
        indent = "  "
        if pretty_text:
            BaseXML.pretty_text(xml_tree, indent=indent)
        xml_string = etree.tostring(xml_tree)
        xml_string = minidom.parseString(xml_string).toprettyxml(indent=indent)
        xml_string = cls.unescape(xml_string)
        return xml_string

    def to_xml(self):
        """NOTICE: This method must be a REENTRANT function, which means
        it should NOT change status or modify any member of `self` object.
        Because other methods may change the construct of the XML tree.
        """
        raise exception.NeedOverrideError()

    def to_str(self, pretty_text=True, remove_first_line=False):
        xml_str = self.to_format_str(self.to_xml(), pretty_text=pretty_text)

        if remove_first_line:
            lines = xml_str.split('\n')
            if len(lines[-1].strip()) == 0:
                # remove last blank line
                lines = lines[1:-1]
            else:
                lines = lines[1:]
            xml_str = '\n'.join(lines)

        return xml_str

    def __str__(self):
        # `remove_first_line=True` is used to remove version tag in the first line
        return self.to_str(remove_first_line=True)



# -*- coding: utf-8 -*-
from __future__ import print_function
from .basexml import BaseXML


class FilterBase(BaseXML):
    """A object represent filter XML node in Karabiner.
    For example, the following XML is a typical filter.

        <only>SKIM</only>
    """
    def __init__(self, *vals, **kwargs):
        """
        :param vals: List[str]
        :param kwargs: Dict
        """
        self.type = kwargs.get('type', 'only')
        self.vals = vals
        self.kwargs = kwargs

    def get_vals(self):
        return self.vals

    def get_tag_name(self):
        header = self.get_class_name().lower().rsplit('filter', 1)[0]
        tag_name = '%s_%s' % (header, self.type)
        return tag_name

    def to_xml(self):
        xml_tree = self.create_tag(self.get_tag_name())
        text = ',\n'.join(self.get_vals())
        self.assign_text_attribute(xml_tree, text)
        return xml_tree

    def __add__(self, another):
        if self.get_tag_name() == another.get_tag_name():
            self.vals += another.vals
            return self
        else:
            tagname1 = self.get_tag_name()
            tagname2 = another.get_tag_name()
            errmsg = "Cannot add %s with %s" % (tagname1, tagname2)
            raise TypeError(errmsg)

    @property
    def id(self):
        return self.get_tag_name(), tuple(sorted(self.get_vals()))

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.id == other.id


class Filter(FilterBase):
    """
    >>> print(Filter('SKIM'))
    <only>SKIM</only>
    >>> print(Filter('SKIM', type='not'))
    <not>SKIM</not>
    """

    def get_tag_name(self):
        return self.type


class ReplacementFilter(Filter):
    pass


class DeviceFilter(FilterBase):
    def get_tag_name(self):
        tag_name = 'device_%s' % self.type
        return tag_name


class DeviceProductFilter(DeviceFilter):
    pass


class DeviceVendorFilter(DeviceFilter):
    pass


class WindowNameFilter(FilterBase):
    pass


class UIElementRoleFilter(FilterBase):
    pass


class InputSourceFilter(FilterBase):
    pass


class ModifierFilter(FilterBase):
    def get_tag_name(self):
        tag_name = 'modifier_%s' % self.type
        return tag_name


if __name__ == "__main__":
    import doctest
    doctest.testmod()
